from __future__ import division
import ctypes
import array
import sys
import platform
import json
import time
import struct

sys.path.append("/home/owner/.local/lib/python3.8/site-packages")
import usb1

"""
libusb 0.1, libusb-win32 documentation:
https://sourceforge.net/p/libusb-win32/wiki/Documentation/

libusb 1.0
https://libusb.info/
https://libusb.sourceforge.io/api-1.0/group__libusb__dev.html#gafee9c4638f1713ca5faa867948878111
"""

VERBOSITY = 2

COMMAND_OUT_EP = 0x01  # To program the data port of the bpi_usb device
COMMAND_IN_EP = 0x81   # Read back log and status information
DATA_OUT_EP = 0x02  # To send data to the bpi_usb device (destined for the ARM or the FPGA)
DATA_IN_EP = 0x82  # To read data from the bpi_usb device (from for the ARM or the FPGA)

"""
MCA ID and USB Product ID:
mca_id byte 0: 0->counter, 1->arm-based MCA, 2->FPGA-based MCA, 3->with eMorpho FPGA;
mca_id byte 1: 1 for PMT, 2 for SiPM
"""

PMT_COUNTER_PID = 0x100
SIPM_COUNTER_PID = 0x200
EMORPHO_PID = 0x6001
PMT3K_MCA_PID = 0x103 # PMT-3000
SIPM3K_MCA_PID = 0x203 # SiPM-3000
PMT2K_MCA_PID = 0x102 # PMT-2000
SIPM2K_MCA_PID = 0x202 # SiPM-2000
PMT1K_MCA_PID = 0x101 # PMT-1000
SIPM1K_MCA_PID = 0x201 # SiPM-1000
PMT_N3K_PID = 0x104 # PMT-3K neutron counter
SIPM_N3K_PID = 0x204 # SiPM-3K neutron counter
COUNTER_3K = 0x005 # Digital pulse counter with ARM and FPGA
HV_ARM = 0x106 # ARM-controlled PMT high voltage supply

pid_list = [EMORPHO_PID, PMT1K_MCA_PID, SIPM1K_MCA_PID, PMT2K_MCA_PID, SIPM2K_MCA_PID, PMT3K_MCA_PID, SIPM3K_MCA_PID,
            PMT_COUNTER_PID, SIPM_COUNTER_PID, PMT_N3K_PID, SIPM_N3K_PID, COUNTER_3K, HV_ARM]  # Valid product ID's
LIBUSB_HAS_GET_DRIVER_NP = sys.platform.startswith('linux')
USE_WIN32 = sys.platform.startswith('win32')  # returns TRUE or FALSE


class bpi_usb:
    """
    A minimal class to count the number of attached MCAs (scan_all) and to claim and open all
    attached MCAs (find_all).
    scan_all() can be used at any time.
    find_all() attempts to claim and open all attached MCAs.  It returns a dictionary of mca objects.
    close_all() can be used to close all MCAs.  But the user will have to discard the mca-object dictionary,
    since its usb handles will now be invalid.
    """
    def __init__(self):
        self.max_packet_size = 64
        self.cmd_out_ep = COMMAND_OUT_EP  # To program the data port of the bpi_usb device
        self.cmd_in_ep = COMMAND_IN_EP  # Read back log and status information
        self.data_out_ep = DATA_OUT_EP  # To send data to the bpi_usb device (destined for the ARM or the FPGA)
        self.data_in_ep = DATA_IN_EP  # To read data from the bpi_usb device (from the ARM or the FPGA)
        self.read_timeout = 1000  # in ms
        self.write_timeout = 1000  # in ms
        self.bpi_vid = 0x1FA4
        self.devices = list()
        self.pids = list()
        self.handles = list()
        self.sernums = list()


    def scan_all(self):
        """Scans for MCA and count them; Does not open an MCA."""
        count = 0
        context = usb1.USBContext()
        device_list = context.getDeviceList(skip_on_error=True)
        for dev in device_list:
            vid = dev.getVendorID()
            pid = dev.getProductID()
            ok = vid == self.bpi_vid
            ok = ok and pid in pid_list
            if ok:
                count += 1
                if VERBOSITY>0:
                    print("PID: {:X}, VID: {:X}".format(pid, vid))
        return count

    def find_all(self):
        """
            Scan the USB bus for all acceptable devices and claim the interface for communication.
        """
        # libusb0.usb_set_debug(4)
        if VERBOSITY>1:
            print("libusb version:", usb1.getVersion())
            print("little endian:", usb1.libusb1.LITTLE_ENDIAN)
        self.handles = list()
        count = 0
        context = usb1.USBContext()
        device_list = context.getDeviceList(skip_on_error=True)
        for dev in device_list:
            vid = dev.getVendorID()
            pid = dev.getProductID()
            ok = vid == self.bpi_vid
            ok = ok and pid in pid_list
            if ok:
                count += 1
                if VERBOSITY>0:
                    print("PID: {:X}, VID: {:X}".format(pid, vid))
                self.devices.append(dev)
                self.pids.append(pid)
                if 0:
                    handle = dev.open()
                    self.handles.append(handle)
                    if LIBUSB_HAS_GET_DRIVER_NP:
                        #handle.detachKernelDriver(0)
                        pass
                    if USE_WIN32:  # This call is necessary for libusb-win32; config_value = 1 always
                        handle.setConfiguration(1)
                    if pid == 0x6001:
                            handle.claimInterface(0)  # device only has interface no. 0
                    else:
                        handle.claimInterface(1)  # device only has interface no. 0

        del context
        print("number of MCA: {}".format(len(self.devices)))
        return self.make_mca_dict()


    def make_mca_dict(self):
        """
            Create a dictionary of mca or eMorpho objects.
            The eMorpho uses an FT245RL USB interface chip and so its methods for I/O and communication with non-volatile memory differ from those devices that have an ARM SAML21 processor with a software-defined USB interface.
        """
        device_name_strings = {
            0x6001: "emorpho",
            0x100: "pmt_counter",
            0x200: "sipm_counter",
            0x101: "pmt_1k",
            0x102: "pmt_2k",
            0x103: "pmt-3k",
            0x201: "sipm-1k",
            0x202: "sipm-2k",
            0x203: "sipm-3k",
            0x104: "pmt-n3k",
            0x204: "sipm-n3k"
        }
        device_type_strings = {
            0x6001: "emorpho",
            0x100: "counter",
            0x200: "counter",
            0x101: "mca_1k",
            0x102: "mca_2k",
            0x103: "mca_3k",
            0x201: "mca_1k",
            0x202: "mca_2k",
            0x203: "mca_3k",
            0x104: "neutron_3k",
            0x204: "neutron_3k"
        }
        mca_dict = {}

        self.sernums = []
        for pid, dev in zip(self.pids, self.devices):
            if VERBOSITY>1:
                print("Device USB address:", dev.getDeviceAddress())
                print("Device speed:", dev.getDeviceSpeed())
                print("Device port number:", dev.getPortNumber())
            # Now that we claimed the interface, we can read the software-defined serial numbers
            if pid == 0x6001:  # eMorpho
                # Reset the FTDI chip

                ser_num = dev.getSerialNumber()

                mca_dict[ser_num] = emorpho()
                mca_dict[ser_num].sn = ser_num
                mca_dict[ser_num].short_sn = ser_num
                mca_dict[ser_num].mca_id = 0x6001
                mca_dict[ser_num].mca_id_str = "0x6001"
                #mca_dict[ser_num].handle = handle
                mca_dict[ser_num].device = dev
                mca_dict[ser_num].set_latency_timer(2)
                self.sernums.append(ser_num)
                mca_dict[ser_num].boot_from_nvmem()
                mca_dict[ser_num].get_adc_sr()
                mca_dict[sn].mca_type_str = device_type_strings[0x6001]
                mca_dict[sn].mca_name_str = device_name_strings[0x6001]
                mca_dict[sn].has_arm = False
                mca_dict[sn].has_fpga = True

            else:  # MCA with ARM
                # Create command-out byte array
                #handle.resetDevice()
                bytes_out = array.array('B', [0] * 64)
                bytes_out[0] = 4   # ARM_READ = 4
                bytes_out[2] = 64  # Read 64 bytes
                handle = dev.open()
                if USE_WIN32:  # This call is necessary for libusb-win32; config_value = 1 always
                    handle.setConfiguration(1)
                handle.claimInterface(1)  # device only has interface no. 1
                handle.bulkWrite(self.cmd_out_ep, bytes_out)
                time.sleep(0.001)
                bytes_in = handle.bulkRead(self.data_in_ep, 64) # Read version registers
                handle.releaseInterface(1)
                del handle
                sn = ''.join(['{:X}'.format(bytes_in[n+8]) for n in range(16)])  # Location of the unique serial number
                mca_dict[sn] = mca()
                #mca_dict[sn].handle = handle
                mca_dict[sn].device = dev
                mca_dict[sn].sn = sn
                mca_id = struct.unpack_from("<1I", bytes_in)[0]
                mca_dict[sn].mca_id = mca_id

                mca_dict[sn].mca_id_str = "0x{:X}".format(mca_id)
                mca_dict[sn].mca_type_str = device_type_strings[mca_id]
                mca_dict[sn].mca_name_str = device_name_strings[mca_id]

                mca_dict[sn].get_adc_sr()
                mca_dict[sn].short_sn = struct.unpack_from("<1I", bytes_in[4: 8])[0]
                arm_hw = struct.unpack_from("<1I", bytes_in[24: 28])[0]
                mca_dict[sn].arm_hw = arm_hw
                mca_dict[sn].hw_id = arm_hw//100000
                mca_dict[sn].date_code = (arm_hw % 100000)//10
                mca_dict[sn].has_arm = mca_id in [0x100, 0x101, 0x102, 0x103, 0x104, 0x200, 0x201, 0x202, 0x203, 0x204]
                mca_dict[sn].has_fpga = mca_id in [0x102, 0x103, 0x104, 0x202, 0x203, 0x204]

                self.sernums.append(sn)
                print("MCA unique ser num: {}; short ser num: {}".format(sn, mca_dict[sn].short_sn))

        return mca_dict


    # close all open devices
    def close(self):
        for device in self.devices:
            device.close()
        self.devices = []  # now an empty list
        self.sernums = []  # now an empty list
        return 0


class mca:
    """
        Class for communication with MCA that have an ARM and a software USB interface.
    """
    def __init__(self):
        self.sn = "0123456789ABCDEF0123456789ABCDEF"
        self.mca_id = 257
        self.mca_id_str = "0x101"
        self.arm_hw = 139423060
        self.hw_id = 1394
        self.date_code = 2306
        self.handle = None
        self.device = None
        self.cmd_out_ep = COMMAND_OUT_EP  # To program the data port of the bpi_usb device
        self.cmd_in_ep = COMMAND_IN_EP  # Read back log and status information
        self.data_out_ep = DATA_OUT_EP  # To send data to the bpi_usb device (destined for the ARM or the FPGA)
        self.data_in_ep = DATA_IN_EP  # To read data from the bpi_usb device (from the ARM or the FPGA)
        self.write_ep = COMMAND_OUT_EP  # The endpoint we write to depends on the command
        self.read_ep = DATA_IN_EP  # # The endpoint we read from depends on the command
        self.read_timeout = 0  # in ms
        self.write_timeout = 0  # in ms
        self.cmd = {}  # The command is a user-supplied dictionary
        self.mem_type = 0,  # 0->RAM, 1->FLASH, 2->factory_reset
        self.num_bytes = 0
        self.bytes_out = array.array('B', [0]*256)
        self.bytes_in = array.array('B', [0]*256)
        self.cmd_in_list = []
        self.cmd_in_dict = {}
        self.cmd_out_list = []
        self.cmd_out_dict = {}
        self.data_in_list = []
        self.data_in_dict = {}
        self.data_out_list = []
        self.data_out_dict = {}
        self.adc_sr = 40e6

    def __repr__(self):
        repr_dict = {
            "sn": self.sn,
            "cmd_in_list": self.cmd_in_list,
            "cmd_out_list": self.cmd_out_list,
            # "cmd_in_dict": self.cmd_in_dict,
            # "cmd_out_dict": self.cmd_out_dict,
            "data_in_list": self.data_in_list,
            # "data_in_dict": self.data_in_dict,
            "data_out_list": self.data_out_list
            # "data_out_dict": self.data_out_dict
        }
        return json.dumps(repr_dict)

    # read byte data from device
    def read_data(self):
        """
        Read data from the device in chunks of 256 bytes or less.  The ARM MCU only buffers data up to 256 bytes.
        The function assumes that the data port on the ARM has been programmed already.

        Returns:
            total number of bytes read

        Side effects:
            Fills self.bytes_in
        """
        handle = self.get_handle()

        chunk_size = 256
        if self.num_bytes <= 0:
            self.bytes_in = array.array('B', [255] * 4)
            return None

        self.num_bytes = int(min(65536, self.num_bytes))  # 64KB is the maximum read size
        num_chunks = self.num_bytes // chunk_size
        frac_length = self.num_bytes % chunk_size
        self.bytes_in = array.array('B', [0] * self.num_bytes)  # Create a receive byte buffer

        total_bytes_read = 0
        for n in range(num_chunks):
            bytes_in = handle.bulkRead(self.read_ep, chunk_size, self.read_timeout)
            bytes_read = len(bytes_in)
            total_bytes_read += bytes_read
            for k in range(bytes_read):
                self.bytes_in[n*chunk_size+k] = bytes_in[k]

        if frac_length > 0:
            bytes_in = handle.bulkRead(self.read_ep, frac_length, self.read_timeout)
            bytes_read = len(bytes_in)
            total_bytes_read += bytes_read
            for k in range(bytes_read):
                self.bytes_in[num_chunks*chunk_size+k] = bytes_in[k]

        self.drop_handle(handle)
        return total_bytes_read

    def wait_100us(self, us):
        us = int(2350*us)
        for n in range(us):
            us+=n

    def write_data(self):
        """
        Write data to the device in chunks of 256 bytes or less.  The ARM MCU only buffers data up to 256 bytes.
        Data to be written to the device are in self.bytes_out

        :return :
            None

        Side effects:
            None
        """

        handle = self.get_handle()

        chunk_size = 256
        if self.num_bytes <= 0:
            return 0

        num_bytes = self.num_bytes
        if self.write_ep == self.cmd_out_ep:  # The command out endpoint requires exactly 64 bytes.
            num_bytes = 64

        num_chunks = num_bytes // chunk_size
        frac_length = num_bytes % chunk_size

        ret = 0
        for n in range(num_chunks):
            handle.bulkWrite(self.write_ep, self.bytes_out[n*chunk_size: (n+1)*chunk_size], self.write_timeout)

        if frac_length > 0:
            handle.bulkWrite(self.write_ep, self.bytes_out[num_chunks*chunk_size: num_chunks*chunk_size+frac_length], self.write_timeout)

        self.drop_handle(handle)
        return ret

    def get_adc_sr(self):
        if self.mca_id not in [0x103, 0x203, 0x104]:
            return
        # a) Prepare to read fpga_results
        num_bytes = 64
        header = (num_bytes << 16) + (2 << 4) + 2

        # Create command-out byte array
        self.write_ep = self.cmd_out_ep
        struct.pack_into("<1I", self.bytes_out, 0, header)  # cmd_out data, to be sent to the device
        struct.pack_into("<30H", self.bytes_out, 4, *([0]*30))  # cmd_out data, to be sent to the device
        self.num_bytes = 64
        self.write_data()  # Write to the command endpoint

        # b) Read the fpga_results
        self.num_bytes = 64
        self.read_data()
        self.data_in_list = struct.unpack_from("<32H", self.bytes_in)
        fpga_ctrl = self.data_in_list
        self.adc_sr = float(fpga_ctrl[6] & 0xFF)*1e6 # ADC speed

    def get_handle(self):
        then = time.time()
        handle = self.device.open()
        if USE_WIN32:  # This call is necessary for libusb-win32; config_value = 1 always
            handle.setConfiguration(1)
        handle.claimInterface(1)  # device only has interface no. 0
        now = time.time()
        #print("dt_get", 1000*(now-then))
        return handle

    def drop_handle(self, handle):
        then = time.time()
        dev = handle.getDevice()
        handle.releaseInterface(1)
        dev.close()
        now = time.time()
        del handle
        #print("dt_drop", 1000*(now-then))


class emorpho:
    """
        Class for communicating with the eMorpho, which has an FT245RL ASIC USB interface.
    """
    def __init__(self):
        self.USB_ENDPOINT_OUT = 0x02
        self.USB_ENDPOINT_IN = 0x81
        self.SIO_SET_LATENCY_TIMER_REQUEST = 0x9
        self.SIO_RESET_REQUEST = 0x0
        self.SIO_RESET_PURGE_RX = 0x1
        self.SIO_RESET_PURGE_TX = 0x2
        self.SIO_RESET_PURGE_RX_TX = 0x3

        self.sn = "eRC0001"
        self.handle = None
        self.write_ep = self.USB_ENDPOINT_OUT  # The endpoint to write to
        self.read_ep = self.USB_ENDPOINT_IN  # # The endpoint to read from
        self.read_timeout = 0  # in ms
        self.write_timeout = 0  # in ms
        self.cmd = {}  # The command is a user-supplied dictionary
        self.bytes_out = array.array('B', [0]*64)
        self.bytes_in = array.array('B', [0]*256)
        self.num_bytes = None
        self.num_bytes_out = 64
        self.cmd_out_list = []
        self.data_in_list = []
        self.data_in_dict = {}
        self.data_out_list = []
        self.data_out_dict = {}
        self.adc_sr = 40e6

        # set latency timer in this device
    def set_latency_timer(self, latency):
        """
        Controls the time out on the device
        :param latency: Timeout in ms; set to 2, except for writing the EEPROM, which needs 0x77
        :return: 0
        """
        handle = self.get_handle()

        latency &= 0xFF  # only lower byte is valid
        data = array.array('B', [0]*1)
        handle.controlWrite(64, self.SIO_SET_LATENCY_TIMER_REQUEST, latency, 0, data)
        self.drop_handle(handle)

        #libusb0.usb_control_msg(self.handle, 64, self.SIO_SET_LATENCY_TIMER_REQUEST, latency, 0, ctypes.c_char_p(0), 0, self.usb_write_timeout)

        return 0

    def purge_tx_buffer(self):
        handle = self.get_handle()
        data = array.array('B', [0]*1)
        handle.controlWrite(64, self.SIO_RESET_REQUEST, self.SIO_RESET_PURGE_TX, 0, data)
        self.drop_handle(handle)
        return
        #return libusb0.usb_control_msg(self.handle, 64, self.SIO_RESET_REQUEST, self.SIO_RESET_PURGE_TX, 0, ctypes.c_char_p(0), 0, self.usb_write_timeout)

    # read byte data from device dev_num
    def read_data(self, bytes_per_datum):
        """
        Read data from the Morpho device to the host
        self.cmd['num_bytes']: number of bytes requested by the host
        bytes_per_datum: 2 or 4
        :return: data array with num_bytes/2 words or num_bytes/4 longs
        """
        self.purge_tx_buffer()  # This empties the transmit FIFO which has 256 bytes of old data.

        handle = self.get_handle()
        read_bytes = int((self.num_bytes//62 + 1)*64)  # Each USB packet contains 2 status bytes
        byte_buf = handle.bulkRead(self.read_ep, read_bytes, self.read_timeout)
        self.drop_handle(handle)

        if bytes_per_datum == 1:
            self.data_in_list = [d for n, d in zip(range(read_bytes), byte_buf) if n & 62]  # Remove modem status bytes
            return None
        bytes_in = array.array('B', [d for n, d in zip(range(read_bytes), byte_buf) if n & 62])  # Remove modem status bytes
        num_bytes = len(bytes_in)

        # Fast conversion from C-array of bytes to python list of integers
        if sys.byteorder == "little":  # This works on little-endian machines
            if bytes_per_datum == 2:
                fmt="<{}H".format(num_bytes//2)

            if bytes_per_datum == 4:
                fmt="<{}I".format(num_bytes//4)

        else:  # big endian
            if bytes_per_datum == 2:
                fmt=">{}H".format(num_bytes//2)

            if bytes_per_datum == 4:
                fmt=">{}I".format(num_bytes//4)

        self.data_in_list = struct.unpack_from(fmt, bytes_in)
        # Do not return more data than expected
        self.data_in_list = self.data_in_list[0:int(self.num_bytes/bytes_per_datum)]
        return len(self.data_in_list)

    def write_data(self):
        """
        Write data to the eMorpho FPGA
        The host filled mca.bytes_out with data
        :return:
        """
        handle = self.get_handle()
        ret = handle.bulkWrite(self.write_ep, self.bytes_out, self.write_timeout)
        self.drop_handle(handle)
        return ret

    def boot_from_nvmem(self):
        """
            The function first reads fpga_results to determine the ADC speed.
            It then reads the nv-mem.  If the data in the nv-mem are valid,
            it will write those fpga_ctrl data into the FPGA RAM.

            We avoid reaching to the API layer, and rather code this functionality
            with what is available in this limited class.
        """
        # Read fpga_results to set adc_sr
        self.num_bytes = 2
        self.bytes_out = array.array('B', [0] * 2)
        struct.pack_into("<1H", self.bytes_out, 0, 0x207)
        self.write_data()
        self.num_bytes = 32
        self.read_data(2)

        self.adc_sr = (self.data_in_list[6] & 0xFF)*1e6

        # Read fpga_ctrl from nv-mem
        # a) Read fpga_ctrl from FPGA
        self.bytes_out = array.array('B', [0] * 2)
        struct.pack_into("<1H", self.bytes_out, 0, 0x007)
        self.write_data()
        self.num_bytes = 32
        self.read_data(2)
        fpga_ctrl_ram = [w for w in self.data_in_list[0:16]]
        fpga_ctrl_ram[13] &= ~0x100  # Clear the sel_lut bit
        fpga_ctrl_ram[15] |= 0x80  # Set the self-clearing read_nv bit

        # b) write the fpga_ctrl back to prepare read from nv_mem to user area
        self.bytes_out = array.array('B', [0] * 34)
        struct.pack_into("<1H", self.bytes_out, 0, 0x003)
        struct.pack_into("<16H", self.bytes_out, 2, *fpga_ctrl_ram)
        self.write_data()
        time.sleep(0.01) # Wait for nv-mem data to be copied into user area

        # c) Read data from user area
        self.bytes_out = array.array('B', [0] * 2)
        struct.pack_into("<1H", self.bytes_out, 0, 0x607)
        self.write_data()
        self.num_bytes = 34 # Check word + 16 fpga_ctrl data
        self.read_data(2)
        fpga_ctrl_nvmem = list(self.data_in_list[1:17])  #self.data_in_list is a tuple

        # d) Check nv-mem for valid content and if valid, copy to fpga_ctrl on the device
        if self.data_in_list[0] == 0x8003:
            fpga_ctrl_nvmem[15] = 0x8000  # Set run bit, but not Program_HV bit
            self.bytes_out = array.array('B', [0] * 34)
            struct.pack_into("<1H", self.bytes_out, 0, 0x003)
            struct.pack_into("<16H", self.bytes_out, 2, *fpga_ctrl_nvmem)
            self.write_data()
            fpga_ctrl = fpga_ctrl_nvmem[0:16]
        else:
            fpga_ctrl = fpga_ctrl_ram[0:16]
        fpga_ctrl[13] &= ~0x100  # Clear the sel_lut bit

        # Now ramp up the high voltage slowly to avoid power surge
        val = fpga_ctrl[7]
        dac_target = ((val & 0xFFF) << 4) + ((val >> 12) & 0xF)
        hv_target = dac_target * 3000/65536
        print("Target HV: ", hv_target)
        for n in range(1, 11):  # Increment HV in 10 steps
            fpga_ctrl[15] = 0x8010  # Set the self-clearing program_hv bit
            dac_val = int((n*dac_target)/10)
            fpga_ctrl[7] = ((dac_val & 0xFFF0) >> 4) + ((dac_val & 0xF) << 12)
            self.bytes_out = array.array('B', [0] * 34)
            struct.pack_into("<1H", self.bytes_out, 0, 0x003)
            struct.pack_into("<16H", self.bytes_out, 2, *fpga_ctrl)
            self.write_data()
            time.sleep(0.1) # Wait for HV to ramp up


    def get_adc_sr(self):
        # a) Read fpga_results
        self.bytes_out = array.array('B', [0] * 2)
        struct.pack_into("<1H", self.bytes_out, 0, 0x207)
        self.write_data()
        self.num_bytes = 32
        self.read_data(2)
        fpga_results = self.data_in_list
        self.adc_sr = float(fpga_results[6] & 0xFF)*1e6 # ADC speed

    def get_handle(self):
        then = time.time()
        handle = self.device.open()
        if USE_WIN32:  # This call is necessary for libusb-win32; config_value = 1 always
            handle.setConfiguration(1)
        handle.claimInterface(0)  # device only has interface no. 0
        now = time.time()
        #print("dt_get", 1000*(now-then))
        return handle

    def drop_handle(self, handle):
        then = time.time()
        dev = handle.getDevice()
        handle.releaseInterface(0)
        dev.close()
        del handle
        now = time.time()
        #print("dt_dropped", 1000*(now-then))

# /home/mike/.local/lib/python3.8/site-packages/usb1
